package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.box.Box
import com.ionspin.kotlin.crypto.box.BoxCorruptedOrTamperedDataException
import com.ionspin.kotlin.crypto.box.crypto_box_NONCEBYTES
import com.ionspin.kotlin.crypto.scalarmult.ScalarMultiplication
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient


object Asymmetric {

    @Serializable
    class Message(
        private val nonce: UByteArray,
        private val encryptedMessage: UByteArray,
        val senderPublicKey: PublicKey? = null,
    ) {

        fun decrypt(recipientKey: SecretKey): UByteArray {
            check(senderPublicKey != null)
            return decryptWithSenderKey(senderPublicKey, recipientKey)
        }

        fun decryptWithSenderKey(senderKey: PublicKey, recipientKey: SecretKey): UByteArray {
            return try {
                Box.openEasy(encryptedMessage, nonce, senderKey.keyBytes, recipientKey.keyBytes)
            } catch (_: BoxCorruptedOrTamperedDataException) {
                throw DecryptionFailedException()
            }
        }
    }

    fun createMessage(
        from: SecretKey, recipient: PublicKey, plainData: UByteArray,
        excludeSenderKey: Boolean = false,
    ): Message {
        val nonce = randomNonce()
        return Message(
            nonce,
            Box.easy(plainData, nonce, recipient.keyBytes, from.keyBytes),
            if (excludeSenderKey) null else from.publicKey
        )
    }

    data class KeyPair(val secretKey: SecretKey, public val senderKey: PublicKey)

    fun generateKeys(): KeyPair {
        val p = Box.keypair()
        val pk = PublicKey(p.publicKey)
        return KeyPair(SecretKey(p.secretKey, pk), pk)
    }

    private fun randomNonce(): UByteArray = randomBytes(crypto_box_NONCEBYTES)

    @Serializable
    class PublicKey(val keyBytes: UByteArray) {
        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is PublicKey) return false

            return keyBytes contentEquals other.keyBytes
        }

        override fun hashCode(): Int {
            return keyBytes.hashCode()
        }
    }

    @Serializable
    class SecretKey(
        val keyBytes: UByteArray,
        @Transient
        val _cachedPublicKey: PublicKey? = null,
    ) {

        private var cachedPublicKey: PublicKey? = _cachedPublicKey

        val publicKey: PublicKey by lazy {
            PublicKey(ScalarMultiplication.scalarMultiplicationBase(keyBytes))
                .also { cachedPublicKey = it }
        }

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is SecretKey) return false

            return keyBytes contentEquals other.keyBytes
        }

        override fun hashCode(): Int {
            return keyBytes.hashCode()
        }
    }

}